// iso_wincmd.cpp : Defines the entry point for the DLL application.
//

#include "stdafx.h"

#include "iso_wincmd.h"

#if defined DEBUG
#define DebugString(x) OutputDebugString( __STR__(__LINE__)": " ),OutputDebugString( x ),OutputDebugString( "\n" )
#define __STR2__(x) #x
#define __STR__(x) __STR2__(x)
#define assert( x ) if( !(x) ) DebugString( "Assertion "#x" in line "__STR__(__LINE__) )
#else
#define DebugString(x)
#define assert( x )
#endif

#ifdef ZeroMemory
#undef ZeroMemory
#endif

#define ZeroMemory( ptr, size ) {for( int i = 0; i < size; i++ ) *((char*)(ptr) + i) = 0;}


static HANDLE heap = 0;

#ifndef DEBUG
static void* malloc( size_t size )
{
    assert( heap );
    return HeapAlloc( heap, HEAP_ZERO_MEMORY, size );
}

static void free( void* block )
{
    assert( heap );
    if( block )
        HeapFree( heap, 0, block );
    return;
}

/*
static void* realloc( void* block, size_t size )
{
    assert( heap );

    if( block && !size )
    {
        HeapFree( heap, 0, block );
        return 0;
    }
    else if( block && size )
        return HeapReAlloc( heap, HEAP_ZERO_MEMORY, block, size );
    else if( !block && size )
        return HeapAlloc( heap, HEAP_ZERO_MEMORY, size );
    else // !block && !size
        return 0;
}
*/
#else
static void* malloc( DWORD size )
{
    assert( heap );

    char* ptr = (char*)HeapAlloc( heap, HEAP_ZERO_MEMORY, size + 12 );
    assert( ptr );
    if( !ptr )
        return 0;
    *(DWORD*)ptr = 0x12345678;
    *(DWORD*)(ptr + 4) = size;
    *(DWORD*)(ptr + size + 8) = 0x12345678;

    return ptr + 8;
}

 void free( void* block )
{
    assert( heap );
    if( block )
    {
        // check for boundary error
        char* ptr = (char*)block;
        if( *(DWORD*)(ptr - 8) != 0x12345678 )
            DebugString( "error before pointer" );
        DWORD size = *(DWORD*)(ptr - 4);
        if( *(DWORD*)(ptr + size) != 0x12345678 )
            DebugString( "error after pointer" );

        HeapFree( heap, 0, ptr - 8 );
    }
    return;
}
/*
static void* realloc( void* block, size_t size )
{
    assert( heap );

    if( block && !size )
    {
        free( block );
        return 0;
    }
    else if( block && size )
    {
        char* p1 = (char*)block;
        char* p2 = (char*)malloc( size );
        DWORD s1 = *(DWORD*)(p1 - 4);
        DWORD s2 = size;
        for( DWORD i = 0; i < min( s1, s2 ); i++ )
            p2[i] = p1[i];
        free( p1 );
        return p2;
    }
    else if( !block && size )
        return malloc( size );
    else // !block && !size
        return 0;
}
*/
#endif //debug

/*
static char* lstrchr( char* str, int chr )
{
    assert( str );
    for( int i = 0; str[i]; i++ )
        if( str[i] == chr )
            return str + i;
    return 0;
}
*/

static char* lstrrchr( char* str, int chr )
{
    assert( str );
    for( int i = lstrlen( str ); i >= 0; i-- )
        if( str[i] == chr )
            return str + i;
    return 0;
}

/*
static char* lstrstr( const char* str1, const char* str2 )
{
    assert( str1 && str2 );

    int l1 = lstrlen( str1 );
    int l2 = lstrlen( str2 );
    for( int i = 0; i < l1 - l2; i++ )
    {
        int j;
        for( j = 0; j < l2; j++ )
           if( str1[i + j] != str2[j] )
               break;
        if( j >= l2 )
            return (char*)str1 + i;
    }
    return 0;
}
*/

static int lmemfind( const char* ptr1, int len1, const char* ptr2, int len2 )
{
    if( len1 < len2 )
        return -1;
    int size = len1 - len2;
    for( int i = 0; i <= size; i++ )
    {
        bool equal = true;
        for( int j = 0; j < len2; j++ )
            if( ptr1[i + j] != ptr2[j] )
            {
                equal = false;
                break;
            }
        if( equal )
            return i;
    }

    return -1;
}

static int lstrcmpn( const char* str1, const char* str2, int len )
{
    if( !len )
        return 0;
    char* b1 = (char*)malloc( len + 2 );
    char* b2 = (char*)malloc( len + 2 );

    lstrcpyn( b1, str1, len + 1 );
    lstrcpyn( b2, str2, len + 1 );

    int res = lstrcmp( b1, b2 );
    free( b1 );
    free( b2 );
    return res;
}


BOOL APIENTRY DllMain( HANDLE hinstDLL, DWORD fdwReason, LPVOID )
{
    if( !hinstDLL )
        return FALSE;

    switch( fdwReason )
    {
        case DLL_PROCESS_ATTACH:
            DebugString( "process attach" );
            //heap = GetProcessHeap();
            heap = HeapCreate( 0, 1024 * 1024, 0 );
            break;
        case DLL_THREAD_ATTACH:
            DebugString( "thread attach" );
            break;
        case DLL_THREAD_DETACH:
            DebugString( "thread detach" );
            break;
        case DLL_PROCESS_DETACH:
            DebugString( "process detach" );
            if( heap )
            {
                HeapDestroy( heap );
                heap = 0;
            }
            break;
    }

    return TRUE;
}

typedef struct VolumeDateTime
{
    unsigned char Year;     // Number of years since 1900
    unsigned char Month;    // Month of the year from 1 to 12
    unsigned char Day;      // Day of the Month from 1 to 31
    unsigned char Hour;     // Hour of the day from 0 to 23
    unsigned char Minute;   // Minute of the hour from 0 to 59
    unsigned char Second;   // second of the minute from 0 to 59
             char Zone;     // Offset from Greenwich Mean Time in
                            // number of 15 minute intervals from
                            // -48(West) to +52(East)
} VolumeDateTime;


#define FATTR_HIDDEN       1
#define FATTR_DIRECTORY    2

typedef struct DirectoryRecord
{
    unsigned char LengthOfDirectoryRecord;  //
    unsigned char ExtendedAttributeRecordLength; // Bytes - this field refers to the
                                            // Extended Attribute Record, which provides
                                            // additional information about a file to
                                            // systems that know how to use it. Since
                                            // few systems use it, we will not discuss
                                            // it here. Refer to ISO 9660:1988 for
                                            // more information.
    __int64       LocationOfExtent;         // This is the Logical Block Number of the
                                            // first Logical Block allocated to the file
    __int64       DataLength;               // Length of the file section in bytes
    VolumeDateTime RecordingDateAndTime;    //
    unsigned char FileFlags;                // One Byte, each bit of which is a Flag:
                                            // bit
                                            // 0 File is Hidden if this bit is 1
                                            // 1 Entry is a Directory if this bit is 1
                                            // 2 Entry is an Associated file is this bit is 1
                                            // 3 Information is structured according to
                                            //   the extended attribute record if this
                                            //   bit is 1
                                            // 4 Owner, group and permissions are
                                            //   specified in the extended attribute
                                            //   record if this bit is 1
                                            // 5 Reserved (0)
                                            // 6 Reserved (0)
                                            // 7 File has more than one directory record
                                            //   if this bit is 1
    unsigned char FileUnitSize;             // This field is only valid if the file is
                                            // recorded in interleave mode
                                            // Otherwise this field is (00)
    unsigned char InterleaveGapSize;        // This field is only valid if the file is
                                            // recorded in interleave mode
                                            // Otherwise this field is (00)
    int VolumeSequenceNumber;               // The ordinal number of the volume in the Volume
                                            // Set on which the file described by the
                                            // directory record is recorded.
    unsigned char LengthOfFileIdentifier;   // Length of File Identifier (LEN_FI)
    unsigned char FileIdentifier[1];
} DirectoryRecord;

typedef struct PathTableRecord
{
    unsigned char LengthOfDirectoryIdentifier;  // Length of Directory Identifier (LEN_DI)
    unsigned char ExtendedAttributeRecordLength;// If an Extended Attribute Record is
                                                // recorded, this is the length in Bytes.
                                                // Otherwise, this is (00)
    unsigned int  LocationOfExtent;             // Logical Block Number of the first Logical
                                                // Block allocated to the Directory
    unsigned short int ParentDirectoryNumber;   // The record number in the Path Table for
                                                // the parent directory of this directory
    unsigned char DirectoryIdentifier[1];       // This field is the same as in the Directory
                                                // Record
} PathTableRecord;


const char CDSignature[] = {'C', 'D', '0', '0', '1'};

typedef struct PrimaryVolumeDescriptor
{
    unsigned char VolumeDescriptorType;
    unsigned char StandardIdentifier[5];        // CD001
    unsigned char VolumeDescriptorVersion;
    unsigned char Ununsed;
    unsigned char SystemIdentifier[32];
    unsigned char VolumeIdentifier[32];
    unsigned char Unused2[8];
    __int64       VolumeSpaceSize;              // Number of logical blocks in the Volume
    unsigned char Unused3[32];
    unsigned int  VolumeSetSize;                // The assigned Volume Set size of the Volume
    unsigned int  VolumeSequenceNumber;         // The ordinal number of the volume in
                                                // the Volume Set
    unsigned int  LogicalBlockSize;             // The size in bytes of a Logical Block
    __int64       PathTableSize;                // Length in bytes of the path table
    unsigned int  LocationOfTypeLPathTable;     // Logical Block Number of first Block allocated
                                                // to the Type L Path Table
    unsigned int  LocationOfOptionalTypeLPathTable; // 0 if Optional Path Table was not recorded,
                                                // otherwise, Logical Block Number of first
                                                // Block allocated to the Optional Type L
                                                // Path Table
    unsigned int  LocationOfTypeMPathTable;     // Logical Block Number of first Block
                                                // allocated to the Type M
    unsigned int  LocationOfOptionalTypeMPathTable; // 0 if Optional Path Table was not
                                                // recorded, otherwise, Logical Path Table,
                                                // Block Number of first Block allocated to the
                                                // Type M Path Table.
    DirectoryRecord DirectoryRecordForRootDirectory; // This is the actual directory record for
                                                // the top of the directory structure
    unsigned char VolumeSetIdentifier[128];     // Name of the multiple volume set of which
                                                // this volume is a member.
    unsigned char PublisherIdentifier[128];     // Identifies who provided the actual data
                                                // contained in the files. acharacters allowed.
    unsigned char DataPreparerIdentifier[128];  // Identifies who performed the actual
                                                // creation of the current volume.
    unsigned char ApplicationIdentifier[128];   // Identifies the specification of how the
                                                // data in the files are recorded.
    unsigned char CopyrightFileIdentifier[37];  // Identifies the file in the root directory
                                                // that contains the copyright notice for
                                                // this volume
    unsigned char AbstractFileIdentifier[37];   // Identifies the file in the root directory
                                                // that contains the abstract statement for
                                                // this volume
    unsigned char BibliographicFileIdentifier[37]; // Identifies the file in the root directory
                                                // that contains bibliographic records.
    unsigned char VolumeCreationDateAndTime[17];// Date and time at which the volume was created
    unsigned char VolumeModificationDateAndTime[17]; // Date and time at which the volume was
                                                // last modified
    unsigned char VolumeExpirationDateAndTime[17]; // Date and Time at which the information in
                                                // the volume may be considered obsolete.
    unsigned char VolumeEffectiveDateAndTime[17];// Date and Time at which the information
                                                // in the volume may be used
    unsigned char FileStructureVersion;         // 1
    unsigned char ReservedForFutureStandardization; // 0
    unsigned char ApplicationUse[512];          // This field is reserved for application use.
                                                // Its content is not specified by ISO-9660.
    unsigned char ReservedForFutureStandardization2[653]; // 0
} PrimaryVolumeDescriptor; // sizeof( PrimaryVolumeDescriptor ) must be 2048


// boot image structures
static const char TORITO[] = "EL TORITO SPECIFICATION";

typedef struct BootRecordVolumeDescriptor
{
    unsigned char BootRecordIndicator;          // Boot Record Indicator, must be 0
    unsigned char StandardIdentifier[5];        // ISO-9660 Identifier, must be "CD001"
    unsigned char VersionOfDescriptor;          // Version of this descriptor, must be 1
    unsigned char BootSystemIdentifier[32];     // Boot System Identifier, must be "EL TORITO SPECIFICATION" padded with 0s.
    unsigned char Unused[32];                   // Unused, must be 0
    unsigned int  BootCatalogPointer;           // Absolute pointer to first sector of Boot Catalog.
    unsigned char Unused2[1973];                // Unused, must be 0.
} BootRecordVolumeDescriptor;

typedef struct ValidationEntry
{
    unsigned char  HeaderID;                    // Header ID, must be 01
    unsigned char  PlatformID;                  // Platform ID
                                                //   0 = 80x86
                                                //   1=Power PC
                                                //   2=Mac
    unsigned short Reserved;                    // Reserved, must be 0
    unsigned char  ID[24];                      // ID string. This is intended to identify the manufacturer/developer of
                                                //   the CD-ROM.
    short int      Checksum;                    // Checksum Word. This sum of all the words in this record should be 0.
    short int      KeyWord;                     // Key word, must be AA55. This value is included in the checksum.
} ValidationEntry;

typedef struct InitialEntry
{
    unsigned char Bootable;                     // Boot Indicator. 88 = Bootable, 00 = Not Bootable
    unsigned char BootMediaType;                // Boot media type. This specifies what media the boot image is intended to
                                                //   emulate in bits 0-3 as follows, bits 4-7 are reserved and must be 0.
                                                //   Bits 0-3 count as follows:
                                                //   0 - No Emulation
                                                //   1 - 1.2 meg diskette
                                                //   2 - 1.44 meg diskette
                                                //   3 - 2.88 meg diskette
                                                //   4 - Hard Disk (drive 80)
                                                //   5-F - Reserved, invalid at this time
    unsigned short LoadSegment;                 // Load Segment. This is the load segment for the initial boot image. If this
                                                //   value is 0 the system will use the traditional segment of 7C0. If this value
                                                //   is non-zero the system will use the specified segment. This applies to x86
                                                //   architectures only. For "flat" model architectures (such as Motorola) this
                                                //   is the address divided by 10.
    unsigned char SystemType;                   // System Type. This must be a copy of byte 5 (System Type) from the
                                                //   Partition Table found in the boot image.
    unsigned char Reserved;                     // Unused, must be 0
    unsigned short SectorCount;                 // Sector Count. This is the number of virtual/emulated sectors the system
                                                //   will store at Load Segment during the initial boot procedure.
    unsigned int  LoadRBA;                      // Load RBA. This is the start address of the virtual disk. CDs use
                                                //   Relative/Logical block addressing.
    unsigned char Reserved2[20];                // Unused, must be 0
} InitialEntry;

typedef struct SectionHeaderEntry
{
    unsigned char HeaderIndicator;              // Header Indicator as follows:
                                                //   90 -Header, more headers follow
                                                //   91 - Final Header
    unsigned char PlatformID;                   // Platform ID:
                                                //   0 = 80x86
                                                //   1 = Power PC
                                                //   2 = Mac
    unsigned short NumberOfSectionEntries;      // Number of section entries following this header
    unsigned char ID[28];                       // ID string. This identifies a section. This string will be checked by BIOS and
                                                //   BOOT software. If the string matches, the section should be scanned for boot
                                                //   images.
} SectionHeaderEntry;

typedef struct SectionEntry
{
    unsigned char BootIndicator;                // Boot Indicator. 88 = Bootable, 00 = Not Bootable
    unsigned char BootMediaType;                // Boot media type. This specifies what media the boot image emulates
                                                //   in bits 0-32.
                                                //   Bits 6 and 7 are specific to the type of system.
                                                //   Bits 0-3 count as follows
                                                //   0 No Emulation
                                                //   1 - 1.2 meg diskette
                                                //   2 - 1.44 meg diskette
                                                //   3 - 2.88 meg diskette
                                                //   4 - Hard Disk (drive 80)
                                                //   5-F - Reserved, invalid at this time
                                                //   bit 4 - Reserved, must be 0
                                                //   bit 5 - Continuation Entry Follows
                                                //   bit 6 - Image contains an ATAPI driver
                                                //   bit 7 - Image contains SCSI drivers
    unsigned short LoadSegment;                 // Load Segment. This is the load segment for the initial boot image.
                                                //   If this value is 0 the system will use the traditional segment
                                                //   of 7C0. If this value is non-zero the system will use the specified
                                                //   segment. This applies to x86 architectures only. For "flat" model
                                                //   architectures (such as Motorola) this is the address divided by 10.
    unsigned char SystemType;                   // System Type. This must be a copy of byte 5 (System Type) from the Partition
                                                //   Table found in the boot image.
    unsigned char Reserved;                     // Unused, must be 0
    unsigned short SectorCount;                 // Sector Count. This is the number of virtual/emulated sectors the system will
                                                //   store at Load Segment during the initial boot procedure.
    unsigned int  LoadRBA;                      // Load RBA. This is the start address of the virtual disk. CDs use
                                                //   Relative/Logical block addressing.
    unsigned char SelectionCriteriaType;        // Selection criteria type. This defines a vendor unique format for bytes 0D-1F.
                                                //   The following formats have currently been assigned:
                                                //   0 - No selection criteria
                                                //   1 - Language and Version Information (IBM)
                                                //   2-FF - Reserved
    unsigned char SelectionCriteria[19];        // Vendor unique selection criteria.
} SectionEntry;

typedef struct SectionEntryExtension
{
    unsigned char ExtensionIndicator;           // Extension Indicator. Must be 44
    unsigned char Bits;                         // Bits 1-4 - Unused
                                                //   5 == 1 - Extension Record follows, 0 - This is final Extension
                                                //   6-7 - Unused
    unsigned char SelectionCriteria[30];        // Vendor unique selection criteria
} SectionEntryExtension;

typedef union CatalogEntry
{
    InitialEntry          Initial;
    SectionHeaderEntry    Header;
    SectionEntry          Entry;
    SectionEntryExtension Extension;
} CatalogEntry;

typedef struct BootCatalog
{
    ValidationEntry Validation;
    CatalogEntry    Entry[2];
} BootCatalog;


typedef struct Directory
{
    char            FilePath[MAX_PATH * 2];
    char            FileName[MAX_PATH];
    DirectoryRecord Record;
} Directory;

typedef struct IsoImage
{
    HANDLE                   hFile;
    PrimaryVolumeDescriptor  VolumeDescriptor;
    DWORD                    DataOffset;
    DWORD                    HeaderSize;
    DWORD                    RealBlockSize;
    Directory*               DirectoryList;
    DWORD                    DirectoryCount;
    DWORD                    Index;
    bool                     Unicode;
    BootCatalog*             BootCatalog;
    DWORD                    BootImageEntries;
} IsoImage;

typedef struct Partition
{
    unsigned char            boot_ind;          /* 0x80 - active */
    unsigned char            head;              /* starting head */
    unsigned char            sector;            /* starting sector */
    unsigned char            cyl;               /* starting cylinder */
    unsigned char            sys_ind;           /* What partition type */
    unsigned char            end_head;          /* end head */
    unsigned char            end_sector;        /* end sector */
    unsigned char            end_cyl;           /* end cylinder */
    unsigned int             start_sect;        /* starting sector counting from 0 */
    unsigned int             nr_sects;          /* nr of sectors in partition */
} Partition;

typedef struct MBR
{
    unsigned char            Loader[0x1be];     // boot loader
    Partition                Partition[4];      // partitions
    unsigned short           Signature;
} MBR;

#define LOWDWORD(x) (DWORD)(x)
#define HIDWORD(x) (DWORD)((x) >> 32)

static DWORD ReadDataByPos( HANDLE file, LONGLONG position, DWORD size, void* data )
{
    assert( file != 0 && file != INVALID_HANDLE_VALUE );
    assert( data );

    LONG low = LOWDWORD( position );
    LONG high = HIDWORD( position );
    
    if( SetFilePointer( file, low, &high, FILE_BEGIN ) != position )
        return 0;

    DWORD read;

    if( ReadFile( file, data, size, &read, 0 ) )
        return read;
    else
        return 0;
}

static LONGLONG GetBlockOffset( IsoImage* image, DWORD block )
{
    return (LONGLONG)(DWORD)block * (WORD)image->RealBlockSize + (image->HeaderSize ? image->HeaderSize : image->DataOffset);
}

static DWORD ReadBlock( IsoImage* image, DWORD block, DWORD size, void* data )
{
    //DebugString( "ReadBlock" );
    assert( image && block && size && data );

    return ReadDataByPos( image->hFile, GetBlockOffset( image, block ), size, data );
}
/*
static bool UpdateImage( IsoImage* image )
{
    if( !image->RealBlockSize )
    {
        char buffer[19];
        if( ReadDataByPos( image->hFile, 0, sizeof( buffer ), buffer ) == sizeof( buffer ) )
        {
            if( !lmemfind( buffer, 12, "\x0\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x0", 12 ) )
            {
                if( buffer[15] == 1 )
                {
                    image->RealBlockSize = 0x930;
                    image->HeaderSize = 0x10;
                    return true;
                }
                else if( buffer[15] == 2 && buffer[18] & 0x20 )
                {
                    image->HeaderSize = 0x18;
                    image->RealBlockSize = 0x930;
                    image->VolumeDescriptor.LogicalBlockSize = 0x914;
                    return true;
                }
                else if( buffer[15] == 2 )
                {
                    image->HeaderSize = 0x18;
                    image->RealBlockSize = 0x930;
                    return true;
                }
            }
        }
        image->RealBlockSize = (WORD)image->VolumeDescriptor.LogicalBlockSize;
    }

    return true;
}
*/
static int ScanBootSections( IsoImage* image, LONGLONG offset, SectionHeaderEntry* firstHeader, SectionEntry* entries )
{
    assert( image );
    assert( firstHeader );

    int images = 0;

    SectionHeaderEntry header = *firstHeader;
    do
    {
        CatalogEntry entry;

        for( int i = 0; i < header.NumberOfSectionEntries; i++, offset += sizeof( entry ) )
        {
            if( ReadDataByPos( image->hFile, offset, sizeof( entry ), &entry ) )
            {
                if( entry.Entry.BootIndicator == 0x88 )
                {
                    if( entries )
                        entries[images] = entry.Entry;
                    images++;
                }

                if( entry.Entry.BootMediaType & (1 << 5) )
                { // next entry is extension
                    for( ; ReadDataByPos( image->hFile, offset, sizeof( entry ), &entry ) &&
                           entry.Extension.ExtensionIndicator == 0x44                     &&
                           entry.Extension.Bits & (5 << 1); offset += sizeof( entry ) );
                }
            }
            else
            {
                DebugString( "can't read SectionEntry" );
                return 0;
            }
        }
    } while( header.HeaderIndicator == 90 );
    return images;
}

static IsoImage* GetImage( const char* filename )
{
    DebugString( "GetImage" );
    if( !filename )
        return 0;

    IsoImage image;
    ZeroMemory( &image, sizeof( image ) );

    image.hFile = CreateFile( filename, GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, 0, 0 );
    if( !image.hFile || image.hFile == INVALID_HANDLE_VALUE )
        return 0;

    assert( sizeof( image.VolumeDescriptor ) == 0x800 ); // check for size of descriptor
    DWORD read;
    
    // Read the VolumeDescriptor (2k) until we find it, but only up to to 1MB
    while( 1 )
    {
        char buffer[sizeof( image.VolumeDescriptor )];
        if( ReadDataByPos( image.hFile, 0x8000 + image.DataOffset, sizeof( buffer ), buffer ) != sizeof( buffer ) )
        {
			// Something went wrong, probably EOF?
			DebugString("Could not read complete VolumeDescriptor block");
			CloseHandle( image.hFile );
			return 0;
        }

        int sig_offset = lmemfind( buffer, sizeof( buffer ), CDSignature, sizeof( CDSignature ) );
        if( sig_offset > 0 )
            image.DataOffset += sig_offset - 1;
        else
        {
            image.DataOffset += sizeof( buffer );
            continue;
        }

        // Try to read a block
        read = ReadDataByPos( image.hFile, 0x8000 + image.DataOffset, 
							  sizeof( image.VolumeDescriptor ), &image.VolumeDescriptor );
		
		if( read != sizeof( image.VolumeDescriptor ) )
		{
			// Something went wrong, probably EOF?
			DebugString("Could not read complete VolumeDescriptor block");
			CloseHandle( image.hFile );
			return 0;
		}

		if( lstrcmpn( (char*)image.VolumeDescriptor.StandardIdentifier, CDSignature, sizeof( CDSignature ) ) == 0 )
		{
			// Found it!
			DebugString("Found it");
			break;
		}
		
		if( image.DataOffset >= 0x100000 )
		{
			// Just to make sure we don't read in a too big file, stop after 1MB.
			DebugString("Reached 1MB without descriptor");
			CloseHandle( image.hFile );
			return 0;
		}
		image.DataOffset += sizeof( image.VolumeDescriptor );
	}

    image.RealBlockSize = (WORD)image.VolumeDescriptor.LogicalBlockSize;

    // detect for next signature CDOO1
    char buffer[10000];
    if( ReadDataByPos( image.hFile, 0x8000 + image.DataOffset + 6, sizeof( buffer ), buffer ) )
    {
        int pos = lmemfind( buffer, sizeof( buffer ), CDSignature, sizeof( CDSignature ) );
        if( pos )
            image.RealBlockSize = pos + 5;
        image.HeaderSize = (0x8000 + image.DataOffset) % image.RealBlockSize;
    }
    
    
    //UpdateImage( &image );

    // check for unicode extention
    DWORD offset = 0x8000 + image.DataOffset + image.RealBlockSize;
	while( !image.Unicode )
    {
        PrimaryVolumeDescriptor descriptor;
        ZeroMemory( &descriptor, sizeof( descriptor ) );
        if( ReadDataByPos( image.hFile, offset, sizeof( descriptor ),
                           &descriptor ) != sizeof( descriptor ) ||
            lstrcmpn( (char*)descriptor.StandardIdentifier, CDSignature, sizeof( CDSignature ) ) )
            break;

        if( descriptor.VolumeDescriptorType == 2                               &&
            (DWORD)descriptor.DirectoryRecordForRootDirectory.LocationOfExtent &&
            (DWORD)descriptor.DirectoryRecordForRootDirectory.DataLength )
        {
            image.Unicode = true;
            image.VolumeDescriptor = descriptor;
			DebugString("image.Unicode = TRUE");
        }
        offset += image.RealBlockSize;
    }

    //UpdateImage( &image );
    
    assert( sizeof( BootRecordVolumeDescriptor ) == 0x800 );
    assert( sizeof( InitialEntry ) == 0x20 );
    assert( sizeof( SectionHeaderEntry ) == 0x20 );
    assert( sizeof( SectionEntry ) == 0x20 );
    assert( sizeof( SectionEntryExtension ) == 0x20 );
    
    BootRecordVolumeDescriptor boot;
    if( ReadDataByPos( image.hFile, 0x8000 + image.DataOffset + image.RealBlockSize, sizeof( boot ), &boot ) )
    { // check for boot image
        if( !boot.BootRecordIndicator &&
            !lstrcmpn( (char*)boot.StandardIdentifier, CDSignature, sizeof( CDSignature ) ) &&
            boot.VersionOfDescriptor == 1 &&
            !lstrcmpn( (char*)boot.BootSystemIdentifier, TORITO, sizeof( TORITO ) ) )
        {
            BootCatalog catalog;
            LONGLONG offset = (LONGLONG)boot.BootCatalogPointer * image.RealBlockSize + image.HeaderSize;
            if( ReadDataByPos( image.hFile, offset, sizeof( catalog ), &catalog ) )
            {
                DebugString( "reading boot catalog" );
                if( (unsigned char)catalog.Validation.HeaderID == (unsigned char)1 &&
                    (short unsigned int)catalog.Validation.KeyWord == (short unsigned int)0xaa55 )
                {
                    unsigned short sum = 0;
                    for( int i = 0; i < sizeof( catalog.Validation ) / sizeof( catalog.Validation.Checksum ); i++ )
                        sum = (unsigned short)(sum + *((unsigned short*)(&catalog.Validation) + i));
                    if( sum )
                    {
                        assert( "checksum is wrong :(" );
                    }
                    else
                    {
                        DebugString( "boot catalog detected" );
                        offset += sizeof( catalog );
                        int images = ScanBootSections( &image, offset, &catalog.Entry[1].Header, 0 );
                        if( catalog.Entry[0].Initial.Bootable == 0x88 )
                            images++;
                        if( images )
                        { // reading images
                            image.BootCatalog = (BootCatalog*)malloc( sizeof( BootCatalog ) +
                                                max( 0, images - 2 ) * sizeof( CatalogEntry ) );
                            assert( image.BootCatalog );
                            if( catalog.Entry[0].Initial.Bootable == 0x88 )
                            {
                                image.BootCatalog->Entry[0] = catalog.Entry[0];
                                ScanBootSections( &image, offset, &catalog.Entry[1].Header,
                                                  &image.BootCatalog->Entry[1].Entry );
                            }
                            else
                                ScanBootSections( &image, offset, &catalog.Entry[1].Header,
                                                  &image.BootCatalog->Entry[0].Entry );
                            image.BootImageEntries = (DWORD)images;
                        }
                    }
                }
                else
                    DebugString( "wrong validation signature" );
            }
            else
                DebugString( "can't read boot catalog from file" );
        }
    }

    IsoImage* pimage = (IsoImage*)malloc( sizeof( *pimage ) );
    assert( pimage );

    *pimage = image;

    return pimage;
}

static bool FreeImage( IsoImage* image )
{
    DebugString( "FreeImage" );
    assert( image );

    if( image->hFile && image->hFile != INVALID_HANDLE_VALUE )
        CloseHandle( image->hFile );

    if( image->DirectoryList )
        free( image->DirectoryList );

    if( image->BootCatalog )
        free( image->BootCatalog );
    
    free( image );

    return true;
}

static bool LoadTree( IsoImage* image, const char* path, DirectoryRecord* root,
                      Directory** dirs, DWORD* count, bool boot = false )
{
    //DebugString( "LoadTree" );
    assert( image && root && count );

    DWORD block = (DWORD)root->LocationOfExtent;
    DWORD size = (DWORD)root->DataLength;

    if( !block || !count )
        return true;

    char* data = (char*)malloc( size );
    assert( data );
    if( !data )
        return false;

    DWORD position = 0;
    for( DWORD k = 0; position < size; position += (WORD)image->VolumeDescriptor.LogicalBlockSize, k++ )
    {
        DWORD s = min( size - position, (WORD)image->VolumeDescriptor.LogicalBlockSize );
        if( ReadBlock( image, block + k, s, data + position ) != s )
        {
            free( data );
            return false;
        }
    }
    
    //if( ReadBlock( image, block, size, data ) != size )
    //{
    //    free( data );
    //    return false;
    //}

    DWORD offset = 0;
    DWORD num = 0;
    while( offset < size - 0x21 )
    {
        while( !data[offset] && offset < size - 0x21 )
            offset++;

        if( offset >= size - 0x21 )
            break;

        DirectoryRecord* record = (DirectoryRecord*)(data + offset);
        Directory directory;
        ZeroMemory( &directory, sizeof( directory ) );
        directory.Record = *record;
        if( image->Unicode && record->LengthOfFileIdentifier > 1 && num > 1 )
        {
            UINT codepage = CP_UTF8;
            int i;
            int length = record->LengthOfFileIdentifier;
            
            for( i = 0; i < record->LengthOfFileIdentifier; i++ )
            {
                if( record->FileIdentifier[i] < 32 )
                {
                    codepage = CP_ACP;
                    length = record->LengthOfFileIdentifier / 2;
                    WCHAR* uname = (WCHAR*)record->FileIdentifier;
                    for( i = 0; i < (record->LengthOfFileIdentifier / 2); i++ )
                        uname[i] = (WORD)((uname[i] >> 8) + (uname[i] << 8));

                    WideCharToMultiByte( codepage, 0, (WCHAR*)record->FileIdentifier, length,
                                         directory.FileName, sizeof( directory.FileName ) - 1, 0, 0 );

                    break;
                }
            }

            if( codepage == CP_UTF8 )
                lstrcpyn( directory.FileName, (char*)record->FileIdentifier, record->LengthOfFileIdentifier + 1 );
            
            if( lstrrchr( directory.FileName, ';' ) )
                *lstrrchr( directory.FileName, ';' ) = 0;

            lstrcat( lstrcat( lstrcpy( directory.FilePath, path ), (path[0]) ? "\\" : "" ), directory.FileName );

            if( dirs )
                (*dirs)[*count] = directory;
            (*count)++;

            if( directory.Record.FileFlags & FATTR_DIRECTORY )
                if( !LoadTree( image, directory.FilePath, &directory.Record, dirs, count ) )
                {
                    free( data );
                    return false;
                }
        }
        else if( !image->Unicode && record->LengthOfFileIdentifier > 0 && num > 1 )
        {
            lstrcpyn( directory.FileName, (char*)record->FileIdentifier, record->LengthOfFileIdentifier + 1);

            if( lstrrchr( directory.FileName, ';' ) )
                *lstrrchr( directory.FileName, ';' ) = 0;

            lstrcat( lstrcat( lstrcpy( directory.FilePath, path ), path[0] ? "\\" : "" ), directory.FileName );

            if( dirs )
                (*dirs)[*count] = directory;
            (*count)++;

            if( directory.Record.FileFlags & FATTR_DIRECTORY )
                if( !LoadTree( image, directory.FilePath, &directory.Record, dirs, count ) )
                {
                    free( data );
                    return false;
                }
        }

        offset += record->LengthOfDirectoryRecord;
        num++;
    }

    if( image->BootImageEntries && boot )
    {
        if( dirs )
        {
            Directory* dir = (*dirs) + (*count);
            ZeroMemory( dir, sizeof( *dir ) );
            lstrcpy( dir->FilePath, lstrcpy( dir->FileName, "boot.images" ) );
            dir->Record.FileFlags = 3;
        }
        (*count)++;
    }

    for( DWORD i = 0; i < image->BootImageEntries && boot; i++ )
    {
        DebugString( "reading boot images" );
        if( dirs )
        {
            Directory* dir = (*dirs) + (*count);
            ZeroMemory( dir, sizeof( *dir ) );
            CatalogEntry* BootEntry = &image->BootCatalog->Entry[i];
            char number[4] = "00";
            number[1] = (char)((i % 10) + '0');
            number[0] = (char)(((i / 10) % 10) + '0');
            
            char mediaType[50];
            const char* floppy = "floppy";
            const char* harddisk  = "harddisk";
            const char* no_emul = "no_emul";
            const char* img = ".";
            //const char* img = ".img";

            switch( BootEntry->Entry.BootMediaType )
            {
            case 0:
                lstrcat( lstrcat( lstrcpy( dir->FileName, no_emul ), img ), number);
                dir->Record.DataLength = BootEntry->Entry.SectorCount * 0x200;
                break;
            case 1:
                lstrcat( lstrcat( lstrcpy( dir->FileName, floppy ), "_1.20."), number );
                dir->Record.DataLength = 0x50 * 0x2 * 0x0f * 0x200;
                break;
            case 2:
                lstrcat( lstrcat( lstrcpy( dir->FileName, floppy ), "_1.44."), number );
                dir->Record.DataLength = 0x50 * 0x2 * 0x12 * 0x200;
                break;
            case 3:
                lstrcat( lstrcat( lstrcpy( dir->FileName, floppy ), "_2.88."), number );
                dir->Record.DataLength = 0x50 * 0x2 * 0x24 * 0x200;
                break;
            case 4:
                {
                    lstrcat( lstrcat( lstrcpy( dir->FileName, harddisk ), img ), number );
                    MBR mbr;
                    if( ReadBlock( image, BootEntry->Entry.LoadRBA, 0x200, &mbr ) )
                        if( mbr.Signature == (unsigned short)0xaa55 )
                            dir->Record.DataLength = (mbr.Partition[0].start_sect + mbr.Partition[0].nr_sects) * 0x200;
                        else
                        {
                            DebugString( "hard disk signature doesn't match" );
                            dir->Record.DataLength = BootEntry->Entry.SectorCount * 0x200;
                        }
                    else
                    {
                        DebugString( "can't read MBR for hard disk emulation" );
                        dir->Record.DataLength = BootEntry->Entry.SectorCount * 0x200;
                    }
                }
                break;
            default:
                lstrcpy( mediaType, ".unknown" );
                dir->Record.DataLength = BootEntry->Entry.SectorCount * 0x200;
                break;
            }
            lstrcat( lstrcpy( dir->FilePath, "boot.images\\" ), dir->FileName );
            dir->Record.LocationOfExtent = BootEntry->Entry.LoadRBA;
        }
        (*count)++;
    }

    free( data );

    return true;
}

static bool ExtractFile( IsoImage* image, Directory* directory, const char* path )
{
    assert( image && directory && path );

    HANDLE file = CreateFile( path, GENERIC_WRITE, FILE_SHARE_READ, 0, CREATE_ALWAYS,
                              FILE_ATTRIBUTE_ARCHIVE, 0 );

    if( !file || file == INVALID_HANDLE_VALUE )
    {
        DebugString( "can't create file" );
        return false;
    }

    DWORD size = (DWORD)directory->Record.DataLength;
    DWORD block = (DWORD)directory->Record.LocationOfExtent;
    DWORD sector = (WORD)image->VolumeDescriptor.LogicalBlockSize;
    char* buffer = (char*)malloc( sector );
    if( !buffer )
    {
        assert( "can't allocate memory for file extract" );
        return false;
    }

    for( ; (int)size > 0; size -= sector, block++ )
    {
        DWORD cur_size = min( sector, size );
        if( ReadBlock( image, block, cur_size, buffer ) != cur_size )
        {
            DebugString( "can't read data" );
            CloseHandle( file );
            free( buffer );
            return false;
        }
        DWORD write;
        WriteFile( file, buffer, cur_size, &write, 0 );
        if( write != min( sector, size ) )
        {
            DebugString( "can't write file" );
            CloseHandle( file );
            free( buffer );
            return false;
        }
    }

    CloseHandle( file );
    free( buffer );

    return true;
}

static tChangeVolProc   ChangeVolProc;
static tProcessDataProc ProcessDataProc;

HANDLE __stdcall OpenArchive( tOpenArchiveData *ArchiveData )
{
    DebugString( "OpenArchive" );
    IsoImage* image = GetImage( ArchiveData->ArcName );
    return image;
}

int __stdcall ReadHeader( HANDLE hArcData, tHeaderData* HeaderData )
{
    //DebugString( "ReadHeader" );

    if( !hArcData || !HeaderData )
        return E_BAD_DATA;

    IsoImage* image = (IsoImage*)hArcData;
    if( !image->DirectoryList )
    {
        DWORD count = 0;
        if( !LoadTree( image, "", &image->VolumeDescriptor.DirectoryRecordForRootDirectory,
                  0, &count, true ) )
            return E_BAD_ARCHIVE;
        if( count )
        {
            image->DirectoryList = (Directory*)malloc( sizeof( Directory ) * count );
            if( !LoadTree( image, "", &image->VolumeDescriptor.DirectoryRecordForRootDirectory,
                      &image->DirectoryList, &image->DirectoryCount, true ) )
                return E_BAD_ARCHIVE;
        }
    }

    if( image->Index >= image->DirectoryCount )
        return E_END_ARCHIVE;

    Directory* directory = image->DirectoryList + image->Index;
    lstrcpy( HeaderData->FileName, directory->FilePath );

    SYSTEMTIME time;
    FILETIME   ftime;
    ZeroMemory( &time, sizeof( time ) );
    time.wYear   = (WORD)(directory->Record.RecordingDateAndTime.Year + 1900);
    time.wMonth  = directory->Record.RecordingDateAndTime.Month;
    time.wDay    = directory->Record.RecordingDateAndTime.Day;
    time.wHour   = directory->Record.RecordingDateAndTime.Hour;
    time.wMinute = directory->Record.RecordingDateAndTime.Minute;
    time.wSecond = directory->Record.RecordingDateAndTime.Second;

    SystemTimeToFileTime( &time, &ftime );
    FileTimeToDosDateTime( &ftime, ((WORD*)&HeaderData->FileTime) + 1, (WORD*)&HeaderData->FileTime );

    HeaderData->FileAttr = FA_ARCHIVE;
    HeaderData->FileAttr |= (directory->Record.FileFlags & FATTR_DIRECTORY) ? FA_DIRECTORY : 0;
    HeaderData->FileAttr |= (directory->Record.FileFlags & FATTR_HIDDEN) ? FA_HIDDEN : 0;

    HeaderData->PackSize = HeaderData->UnpSize = (int)directory->Record.DataLength;

    return 0;
}

int __stdcall ProcessFile( HANDLE hArcData, int Operation, char* DestPath, char* DestName )
{
    //DebugString( "ProcessFile" );
    if( !hArcData )
        return E_BAD_DATA;

    IsoImage* image = (IsoImage*)hArcData;

    bool result = true;
    if( Operation == PK_EXTRACT )
    {
        DebugString( "Extracting" );
        char path[MAX_PATH * 2];
        path[0] = 0;
        if( DestPath )
            lstrcpy( path, DestPath );

        if( DestName )
            lstrcat( lstrcat( path, path[0] ? "\\" : "" ), DestName );

        result = ExtractFile( image, &image->DirectoryList[image->Index], path );
        if( !ProcessDataProc( image->DirectoryList[image->Index].FileName,
                              (int)image->DirectoryList[image->Index].Record.DataLength ) )
            result = false;
    }

    image->Index++;
    return result ? 0 : E_EWRITE;
}

int __stdcall CloseArchive( HANDLE hArcData )
{
    DebugString( "CloseArchive" );
    FreeImage( (IsoImage*)hArcData );
    return 0;
}

void __stdcall SetChangeVolProc( HANDLE /*hArcData*/, tChangeVolProc pChangeVolProc )
{
    ChangeVolProc = pChangeVolProc;
}

void __stdcall SetProcessDataProc( HANDLE /*hArcData*/, tProcessDataProc pProcessDataProc )
{
    ProcessDataProc = pProcessDataProc;
}

BOOL __stdcall CanYouHandleThisFile( char* FileName )
{
    if( !FileName )
        return FALSE;

    IsoImage* image = GetImage( FileName );

    if( !image )
        return FALSE;
    else
    {
        FreeImage( image );
        return TRUE;
    }
}

int __stdcall GetPackerCaps()
{
    return PK_CAPS_BY_CONTENT | PK_CAPS_SEARCHTEXT;
}



